#include "AY.h"
#include <memory.h>
#include <stdlib.h>

/* liberated from FUSE */
const int Levels[16]=
{
	0x0000 >> 2, 0x0385 >> 2, 0x053D >> 2, 0x0770 >> 2,
	0x0AD7 >> 2, 0x0FD5 >> 2, 0x15B0 >> 2, 0x230C >> 2,
	0x2B4C >> 2, 0x43C1 >> 2, 0x5A4B >> 2, 0x732F >> 2,
	0x9204 >> 2, 0xAFF1 >> 2, 0xD921 >> 2, 0xFFFF >> 2
};

#define TONE_BIT	0x400000
#define TONE_SHIFT	22

#define CLOCK_RATE 3545400

CAY::CAY()
{
	CReg = 0;
	TotalTime = 0;
	AudioReadPtrMutex = NULL;

	/* setup noise and patterns */
	int c = 1024;
	while(c--)
	{
		NoiseWave[c] = (rand() << 1) / RAND_MAX;
	}

	EnvelopePatterns[1][0] = 0;
	EnvelopePatterns[3][0] =
	EnvelopePatterns[4][0] = 15;

	c = 16;
	while(c--)
	{
		EnvelopePatterns[0][c] = c;
		EnvelopePatterns[1][c+1] = c^0xf;

		EnvelopePatterns[2][c] = c;
		EnvelopePatterns[2][c+16] = c^0xf;

		EnvelopePatterns[3][c+1] = c;
		EnvelopePatterns[4][c] = c^0xf;
	}

	c = 16;
	while(c--)
	{
		EnvelopePatterns[0][c+16] = EnvelopePatterns[0][c];
		EnvelopePatterns[1][c+16] = EnvelopePatterns[1][c];
		EnvelopePatterns[3][c+16] = EnvelopePatterns[3][c];
		EnvelopePatterns[4][c+16] = EnvelopePatterns[4][c];
	}

	EnvelopePattern = EnvelopePatterns[0];
	EnvelopeAdd = 0;
	EnvelopePos = 0;

	c = 5;
	while(c--)
	{
		int ic = 32;
		while(ic--)
			EnvelopePatterns[c][ic] = Levels[EnvelopePatterns[c][ic]];
	}

	/* initialise audio */
	SDL_AudioSpec WAudioSpec;

	WAudioSpec.freq = 44100;
	
#if SDL_BYTEORDER == SDL_LIL_ENDIAN
	WAudioSpec.format = AUDIO_S16;
#else
	WAudioSpec.format = AUDIO_S16MSB;
#endif	
	WAudioSpec.channels = 1;
	WAudioSpec.samples = 4096;
	WAudioSpec.callback = AudioUpdateFunctionHelper;
	WAudioSpec.userdata = this;

	AudioWritePtr = AudioReadPtr = 0;
	AudioProcessTime = AudioMask = 0;
	AudioBuffer[CULA_AUDIOEVENT_LENGTH-1].ClockTime = 0;

	if(SDL_OpenAudio(&WAudioSpec, &AudioSpec) >= 0)
	{
		fprintf(stderr, "Audio open: %d\n", AudioSpec.freq);
		AudioEnabled = true;
		AudioReadPtrMutex = SDL_CreateMutex();
		AudioNumerator = (Uint32)(((float)CLOCK_RATE / (float)(AudioSpec.freq << 4)) * (1 << TONE_SHIFT));
		SDL_PauseAudio(0);

		int c = 3;
		while(c--)
		{
			Channels[c].Pitch = Channels[c].ToneMask = Channels[c].NoiseMask = 0;
			Channels[c].Volume = 0;
			Channels[c].VolPtr = &Channels[c].Volume;
			Channels[c].SamplePos = Channels[c].SampleAdd = 0;
		}
	}
	else
	{
		fprintf(stderr, "Couldn't open audio\n");
		AudioEnabled = false;
	}
}

CAY::~CAY()
{
	SDL_CloseAudio();
	if(AudioReadPtrMutex) SDL_DestroyMutex(AudioReadPtrMutex);
}

void CAY::Write(CZ80 &CPU, Uint16 Addr, Uint8 Value8, Uint64 Value64, Uint32 Time)
{
	switch(Addr)
	{
		case 0xfffd:
			CReg = Value8&15;
		break;

		case 0xbffd:
			switch(CReg)
			{
				case 1: case 3: case 5: 
				case 13: Value8 &= 0x0f; break;

				case 8: case 9: case 10:
				case 6: Value8 &= 0x1f; break;
			}
			Registers[CReg] = Value8;

			AudioBuffer[AudioWritePtr].Register = CReg;
			AudioBuffer[AudioWritePtr].Value = Value8;
			WriteAudioEvent(Time);
		break;

		case CAY_TOGGLEADDR:
			AudioBuffer[AudioWritePtr].Register = 16;
			AudioBuffer[AudioWritePtr].Value = Value8;
			WriteAudioEvent(Time);
		break;
	}
}

void CAY::Read(CZ80 &CPU, Uint16 Addr, Uint8 &Value8, Uint64 &Value64, Uint32 Time)
{
	Value8 = Registers[CReg];
}

Uint32 CAY::Update(Uint32 cycles, bool)
{
	TotalTime += cycles;

	if(AudioEnabled)
	{
		/* generate NOP event to keep sound system alert and awake */
		AudioBuffer[AudioWritePtr].Register = 17;
		WriteAudioEvent(TotalTime);

		/* force resync if samplediff reveals more than an eighth of a second */
		Uint32 OldTime = AudioBuffer[AudioReadPtr&(CULA_AUDIOEVENT_LENGTH-1)].ClockTime;
		Uint32 BigDiff = TotalTime - OldTime;

		if(BigDiff > (CLOCK_RATE >> 3))
		{
			SDL_mutexP(AudioReadPtrMutex);
			while(BigDiff > (CLOCK_RATE >> 3))
			{
				EnactAudioEvent();
				BigDiff = TotalTime - AudioBuffer[AudioReadPtr&(CULA_AUDIOEVENT_LENGTH-1)].ClockTime;
			}
			SDL_mutexV(AudioReadPtrMutex);
		}
	}
	return 0;
}

void CAY::FixChannel(int c)
{
	Channels[c].SampleAdd = AudioNumerator / (Channels[c].Pitch+1);

	if(Channels[c].SampleAdd > TONE_BIT)
	{
		Channels[c].SampleAdd = 0;
		Channels[c].SamplePos = TONE_BIT;
	}
}

void CAY::FixEnvelope()
{
	EnvelopeAdd = (AudioNumerator >> 4) / (EnvelopePitch+1);
}


void CAY::EnactAudioEvent()
{
	/* enact top event */
	switch(AudioBuffer[ AudioReadPtr ].Register)
	{
		case 0:	Channels[0].Pitch = (Channels[0].Pitch&0xf00) | AudioBuffer[ AudioReadPtr ].Value; FixChannel(0); break;
		case 1: Channels[0].Pitch = (Channels[0].Pitch&0x0ff) | (AudioBuffer[ AudioReadPtr ].Value << 8); FixChannel(0); break;

		case 2:	Channels[1].Pitch = (Channels[1].Pitch&0xf00) | AudioBuffer[ AudioReadPtr ].Value; FixChannel(1); break;
		case 3: Channels[1].Pitch = (Channels[1].Pitch&0x0ff) | (AudioBuffer[ AudioReadPtr ].Value << 8); FixChannel(1); break;

		case 4:	Channels[2].Pitch = (Channels[2].Pitch&0xf00) | AudioBuffer[ AudioReadPtr ].Value; FixChannel(2); break;
		case 5: Channels[2].Pitch = (Channels[2].Pitch&0x0ff) | (AudioBuffer[ AudioReadPtr ].Value << 8); FixChannel(2); break;

		case 6:	NoiseAdd = AudioNumerator / (AudioBuffer[ AudioReadPtr ].Value+1);	break;

		case 7:
			Channels[0].ToneMask = (AudioBuffer[ AudioReadPtr ].Value&0x01) ? 0x0000 : 0xffff;
			Channels[1].ToneMask = (AudioBuffer[ AudioReadPtr ].Value&0x02) ? 0x0000 : 0xffff;
			Channels[2].ToneMask = (AudioBuffer[ AudioReadPtr ].Value&0x04) ? 0x0000 : 0xffff;
			Channels[0].NoiseMask = (AudioBuffer[ AudioReadPtr ].Value&0x08) ? 0x0000 : 0xffff;
			Channels[1].NoiseMask = (AudioBuffer[ AudioReadPtr ].Value&0x10) ? 0x0000 : 0xffff;
			Channels[2].NoiseMask = (AudioBuffer[ AudioReadPtr ].Value&0x20) ? 0x0000 : 0xffff;
		break;

#define SetVolume(n)\
	if(AudioBuffer[ AudioReadPtr ].Value&0x10)\
		Channels[n].VolPtr = &EnvelopeVolume;\
	else\
	{\
		Channels[n].VolPtr = &Channels[n].Volume;\
		Channels[n].Volume = Levels[AudioBuffer[ AudioReadPtr ].Value];\
	}

		case 8:		SetVolume(0); break;
		case 9:		SetVolume(1); break;
		case 10:	SetVolume(2); break;

		case 11:	EnvelopePitch = (EnvelopePitch&0xff00) | AudioBuffer[ AudioReadPtr ].Value; FixEnvelope(); break;
		case 12:	EnvelopePitch = (EnvelopePitch&0x00ff) | (AudioBuffer[ AudioReadPtr ].Value << 8); FixEnvelope(); break;
		case 13:
			/*
				0, 9 - decay, off			- Tab1, 15 to 0
				4, 15 - attack, off			- Tab2, 15 to 0

				8 - repeated decay			- Tab1, 15 repeat
				10 - decay, attack, repeat	- Tab3, 15 repeat
				11 - decay, hold			- Tab4

				12 - repeated attack		- Tab2, 16 repeat
				14 - attack, decay, repeat	- Tab3, 31 repeat
				13 - attack, hold			- Tab5

				Tab1 = 0, 1, ... 15, [repeat]
				Tab2 = 0, 15 ... 1, [repeat]
				Tab3 = 0, 1 ...15, 15, ..., 0
			*/
//				printf("envelope %d\n", AudioBuffer[ AudioReadPtr ].Value);
				switch(AudioBuffer[ AudioReadPtr ].Value)
				{
					case 0: case 1: case 2: case 3: case 9:
						EnvelopeMask = 0xffffffff;
						EnvelopePattern = EnvelopePatterns[0];
						EnvelopePos = 15 << TONE_SHIFT;
					break;

					case 4: case 5: case 6: case 7: case 15:
						EnvelopeMask = 0xffffffff;
						EnvelopePattern = EnvelopePatterns[1];
						EnvelopePos = 15 << TONE_SHIFT;
					break;

					case 8: 
						EnvelopeMask = ~(~0xf << TONE_SHIFT);
						EnvelopePattern = EnvelopePatterns[0];
						EnvelopePos = 15 << TONE_SHIFT;
					break;
					case 12:
						EnvelopeMask = ~(~0x1f << TONE_SHIFT);
						EnvelopePattern = EnvelopePatterns[1];
						EnvelopePos = 16 << TONE_SHIFT;
					break;

					case 10:
						EnvelopeMask = ~(~0x1f << TONE_SHIFT);
						EnvelopePattern = EnvelopePatterns[2];
						EnvelopePos = 15 << TONE_SHIFT;
					break;
					case 14:
						EnvelopeMask = ~(~0x1f << TONE_SHIFT);
						EnvelopePattern = EnvelopePatterns[2];
						EnvelopePos = 31 << TONE_SHIFT;
					break;

					case 11:
						EnvelopeMask = 0xffffffff;
						EnvelopePattern = EnvelopePatterns[3];
						EnvelopePos = 16 << TONE_SHIFT;
					break;
					case 13:
						EnvelopeMask = 0xffffffff;
						EnvelopePattern = EnvelopePatterns[4];
						EnvelopePos = 16 << TONE_SHIFT;
					break;
				}

		case 16:
			SpeakerLevel = (AudioBuffer[ AudioReadPtr ].Value ? 64 : 0) << 8;
		break;

	}

	AudioReadPtr = (AudioReadPtr+1)&(CULA_AUDIOEVENT_LENGTH-1);
}

void CAY::AudioUpdateFunctionHelper(void *tptr, Uint8 *TargetBuffer, int TargetLength)
{
	((CAY *)tptr)->AudioUpdateFunction(NULL, TargetBuffer, TargetLength);
}

void CAY::AudioUpdateFunction(void *tptr, Uint8 *TargetBuffer, int TargetLength)
{
	SDL_mutexP(AudioReadPtrMutex);

	int SPtr = 0;
	TargetLength >>= 1;
	while(SPtr < TargetLength)
	{
		while(
			( AudioReadPtr != AudioWritePtr) &&
			( AudioProcessTime >= AudioBuffer[ AudioReadPtr ].SampleDiff)
		)
		{
			EnactAudioEvent();
			AudioProcessTime = 0;
		}

		Uint32 SamplesToWrite;
		if(AudioReadPtr == AudioWritePtr)
		{
			SamplesToWrite = TargetLength-SPtr;
		}
		else
		{
			SamplesToWrite = AudioBuffer[ AudioReadPtr ].SampleDiff - AudioProcessTime;

			if(SamplesToWrite > (unsigned)(TargetLength-SPtr))
				SamplesToWrite = TargetLength-SPtr;
		}

		/*
			BEHOLD: the clauseless AY sound emulation inner loop (*ahem* which works only
			if your computer allows signed shifts)
		*/

		AudioProcessTime += SamplesToWrite;
		while(SamplesToWrite--)
		{
#define Generator(n)\
		(\
			*Channels[n].VolPtr -\
			(\
				(((Channels[n].SamplePos&TONE_BIT) ? 0 : *Channels[n].VolPtr )&Channels[n].ToneMask) |\
				((NoiseWave[(NoisePos >> TONE_SHIFT)&1023] ? 0 : *Channels[n].VolPtr)&Channels[n].NoiseMask)\
			)\
		)

			EnvelopeVolume = EnvelopePattern[EnvelopePos >> TONE_SHIFT];
			((Uint16 *)TargetBuffer)[SPtr++] = -32768 + Generator(0) + Generator(1) + Generator(2) + SpeakerLevel;

			Channels[0].SamplePos += Channels[0].SampleAdd;
			Channels[1].SamplePos += Channels[1].SampleAdd;
			Channels[2].SamplePos += Channels[2].SampleAdd;
			NoisePos += NoiseAdd;
			EnvelopePos = (EnvelopePos-EnvelopeAdd)&EnvelopeMask;
			EnvelopePos &= ~(EnvelopePos >> 31);
		}
	}

	SDL_mutexV(AudioReadPtrMutex);
}

void CAY::WriteAudioEvent(Uint32 TimeStamp)
{
	AudioBuffer[AudioWritePtr].ClockTime = TimeStamp;

	Uint32 Difference = TimeStamp - AudioBuffer[(AudioWritePtr-1)&(CULA_AUDIOEVENT_LENGTH-1)].ClockTime;

	AudioBuffer[AudioWritePtr].SampleDiff = (Difference*AudioSpec.freq + AudioBuffer[(AudioWritePtr-1)&(CULA_AUDIOEVENT_LENGTH-1)].Remainder) / CLOCK_RATE;
	AudioBuffer[AudioWritePtr].Remainder = (Difference*AudioSpec.freq + AudioBuffer[(AudioWritePtr-1)&(CULA_AUDIOEVENT_LENGTH-1)].Remainder) % CLOCK_RATE;
	AudioWritePtr = (AudioWritePtr+1)&(CULA_AUDIOEVENT_LENGTH-1);
}
