#include "Snapshots.h"
#include "z80.h"
#include <stdio.h>
#include <string.h> 
#include "ULA.h"
#include "Screen.h"

#define Read16(f, v)\
	v.b.b0 = fgetc(f);\
	v.b.b1 = fgetc(f);

/*
	LoadSNA loads a 48kb or 128kb SNA. Returns false if load unsucessful, true otherwise
*/
bool LoadSNA(CZ80 &CPU, char *fname, unsigned int Timestamp)
{
	FILE *sna = fopen(fname, "rb");
	if(!sna) return false;

	RegisterState NS;

	NS.IReg = fgetc(sna);

	Read16(sna, NS.HL2);
	Read16(sna, NS.DE2);
	Read16(sna, NS.BC2);
	Read16(sna, NS.AF2);
	Read16(sna, NS.HL);
	Read16(sna, NS.DE);
	Read16(sna, NS.BC);
	Read16(sna, NS.IY);
	Read16(sna, NS.IX);

	Uint8 Temp8 = fgetc(sna);
	NS.IFF1 = NS.IFF2 = (Temp8&4) ? true : false;

	NS.RReg = fgetc(sna);
	Read16(sna, NS.AF);

	Read16(sna, NS.SP);
	NS.IMode = fgetc(sna);

	Temp8 = fgetc(sna);	/* border colour */
	CPU.WritePort(0xfe, Temp8, Timestamp);

	/* read RAM contents */
	Uint8 PagedRam[16384];
	fread(CPU.GetPtr8(5+2), 1, 16384, sna);
	fread(CPU.GetPtr8(2+2), 1, 16384, sna);
	fread(PagedRam, 1, 16384, sna);

	/* read PC / check if a 128kb sna */
	NS.PC.b.b0 = fgetc(sna);
	if(feof(sna))
	{
		/* set for 48kb mode */
		CPU.WritePort(0x7ffd, 4 | 0x10 | 0x20, Timestamp);
		memcpy(CPU.GetPtr8(4+2), PagedRam, 16384);

		/* pop PC */
		CPU.ReadMemory(NS.SP.a, NS.PC.b.b0); NS.SP.a++;
		CPU.ReadMemory(NS.SP.a, NS.PC.b.b1); NS.SP.a++;

		fclose(sna);
		CPU.SetState(NS);
		CPU.DumpMultiplexedCopy();
		return true;
	}

	NS.PC.b.b1 = fgetc(sna);

	/* port 7ffd setting */
	Temp8 = fgetc(sna);
	CPU.WritePort(0x7ffd, Temp8, Timestamp);

	/* copy third bank of RAM to correct place */
	memcpy(CPU.GetPtr8((Temp8&7)+2), PagedRam, 16384);

	/* skip TRDOS byte */
	fseek(sna, 1, SEEK_CUR);

	/* read rest of RAM */
	int PageTable[] = {0, 1, 3, 4, 6, 7};
	for(int pagec = 0; pagec < 6; pagec++)
	{
		if((Temp8&7) != PageTable[pagec])
		{
			fread(CPU.GetPtr8(PageTable[pagec]+2), 1, 16384, sna);
		}
	}

	CPU.SetState(NS);
	CPU.DumpMultiplexedCopy();

	fclose(sna);
	return true;
}

/*
	LoadZ80 loads a 48kb or 128kb Z80. Returns false if load unsucessful, true otherwise
*/
FILE *z80;
bool Compressed, Version1;
Uint8 Byte, Repeat;
Uint16 DataLength;

void InitialiseZ80Read(Uint16 DL = 0xffff)
{
	Repeat = 0;
	DataLength = DL;
}

Uint8 Z80ReadByte()
{
	if(!DataLength) return 0xff;

	if(Repeat)
	{
		Repeat--;
		return Byte;
	}

	Uint8 Buf[2];
	Buf[0] = fgetc(z80); DataLength--;
	if(Buf[0] == 0xed)
	{
		Buf[1] = fgetc(z80);
		if(Buf[1] != 0xed)
		{
			ungetc(Buf[1], z80);
			return Buf[0];
		}
		DataLength-= 3;

		Repeat = fgetc(z80) - 1;
		return Byte = fgetc(z80);
	}

	return Buf[0];
}

void Z80Read( Uint8 *memory, int length)
{
	if(!Compressed)
	{
		fread(memory, 1, length, z80);
		return;
	}

	while(length--)
	{
		*memory++ = Z80ReadByte();
	}
}

bool LoadZ80(CZ80 &CPU, char *fname, unsigned int Timestamp)
{
	z80 = fopen(fname, "rb");
	if(!z80) return false;

	RegisterState NS;

	NS.AF.b.b1 = fgetc(z80);
	NS.AF.b.b0 = fgetc(z80);

	Read16(z80, NS.BC);
	Read16(z80, NS.DE);
	Read16(z80, NS.PC);
	Read16(z80, NS.SP);

	NS.IReg = fgetc(z80);
	NS.RReg = fgetc(z80) & 0x7f;
	Uint8 RHigh = fgetc(z80); if(RHigh == 0xff) RHigh = 1;
	NS.RReg |= RHigh << 7;

	/* border colour */
	CPU.WritePort(0xfe, (RHigh >> 1)&7, Timestamp);
	Compressed = (RHigh & 0x20) ? true : false;

	Read16(z80, NS.DE);
	Read16(z80, NS.BC2);
	Read16(z80, NS.DE2);
	Read16(z80, NS.HL2);
	NS.AF2.b.b1 = fgetc(z80);
	NS.AF2.b.b0 = fgetc(z80);
	Read16(z80, NS.IY);
	Read16(z80, NS.IX);

	NS.IFF1 = fgetc(z80) ? true : false;
	NS.IFF2 = fgetc(z80) ? true : false;

	NS.IMode = fgetc(z80)&3;

	if(NS.PC.a)
	{
		/* old fashioned z80 - 48kb only */
		Version1 = true;

		/* set for 48kb mode */
		CPU.WritePort(0x7ffd, 4 | 0x10 | 0x20, Timestamp);

		InitialiseZ80Read();
		Z80Read( CPU.GetPtr8(5+2), 16384);
		Z80Read( CPU.GetPtr8(2+2), 16384);
		Z80Read( CPU.GetPtr8(4+2), 16384);
	}
	else
	{
		/* version 2 or above z80 */
		Version1 = false;
		Compressed = true;

		RegPair HeaderLength;
		Read16(z80, HeaderLength);
		Read16(z80, NS.PC);

		Uint8 Hardware = fgetc(z80);
		bool e48kb = false;

		if(HeaderLength.a == 23)
		{
			/* version 2.x file */
			switch(Hardware)
			{
				default:
					fclose(z80);
				return false;

				case 0:
					/* set for 48kb mode */
					CPU.WritePort(0x7ffd, 4 | 0x10 | 0x20, Timestamp);
					e48kb = true;
				break;

				case 3: break; //128kb mode
			}
		}
		else
		{
			/* version 3.0 file */
			switch(Hardware)
			{
				default:
					fclose(z80);
				return false;

				case 0:
					/* set for 48kb mode */
					CPU.WritePort(0x7ffd, 4 | 0x10 | 0x20, Timestamp);
					e48kb = true;
				break;

				case 4: break; //128kb mode
			}
		}

		Uint8 Output7ffd = fgetc(z80);
		CPU.WritePort(0x7ffd, Output7ffd, Timestamp);

		fseek(z80, 2, SEEK_CUR); //skip IFF1 & z80 specific emulation stuff

		/* AY state */
		Uint8 Outputfffd = fgetc(z80);
		int c = 16;
		while(c--)
		{
			CPU.WritePort(0xfffd, c, Timestamp);
			CPU.WritePort(0xbffd, fgetc(z80), Timestamp);
		}
		CPU.WritePort(0xfffd, Outputfffd, Timestamp);

		fseek(z80, HeaderLength.a - 23, SEEK_CUR); //skip other useless stuff

		/* parse memory blocks */
		Uint8 Map128kb[12] =
		{
			0, 0xff, 1,
			0+2, 1+2, 2+2, 3+2,
			4+2, 5+2, 6+2, 7+2,
			0xff
		};
		Uint8 Map48kb[12] =
		{
			1, 0xff, 0xff,
			0xff, 2+2, 4+2, 0xff,
			0xff, 5+2, 0xff, 0xff,
			0xff
		};
		Uint8 *CMap = e48kb ? Map48kb : Map128kb;
		while(1)
		{
			RegPair DataLen;
			Read16(z80, DataLen);
			long endpos = ftell(z80) + ((DataLen.a == 0xffff) ? 16384 : DataLen.a) + 1;
			if(feof(z80)) break;

			Uint8 PageNo = fgetc(z80);
			InitialiseZ80Read(DataLen.a);

			if((PageNo < 12) && CMap[PageNo])
			{
				if(DataLen.a == 0xffff)
					fread(CPU.GetPtr8(CMap[PageNo]), 1, 16384, z80);
				else
				{
					Z80Read(CPU.GetPtr8(CMap[PageNo]), 16384);
					Z80Read((Uint8 *)CPU.GetPtr64(CMap[PageNo]), 16384*8);
				}
			}

			/* special page: palette */
			if(PageNo == 0xff)
			{
				Uint8 Palette[128*256*3 + 256*3];
				Z80Read(Palette, 128*256*3 + 256*3);
				SCR->SetMultiPalette(Palette);
				SCR->SetFlags( SCR->GetFlags() | CSF_MULTIPLEXED);
			}

			fseek(z80, endpos, SEEK_SET);
		}
	}

	fclose(z80);
	CPU.SetState(NS);
	return true;
}
