#ifndef __z80_H
#define __z80_H

#include "SDL.h"
#include "SDL_thread.h"

union RegPair
{
	Uint16 a;

#if SDL_BYTEORDER == SDL_LIL_ENDIAN
	struct
	{
		Uint8 b0, b1;
	} b;
#else
	struct
	{
		Uint8 b1, b0;
	} b;
#endif
};

struct RegisterState
{
	bool IFF1, IFF2;
	Uint8 IMode, RReg, IReg;
	RegPair AF, BC, DE, HL, SP, BC2, DE2, HL2, AF2, IY, IX, PC;

	/* large registers */
	Uint64 ABig, ABig2;
	Uint64 BBig, CBig, BBig2, CBig2;
	Uint64 DBig, EBig, DBig2, EBig2;
	Uint64 HBig, LBig, HBig2, LBig2;
};

class CZ80;

class CZ80IO
{
	public:
		virtual ~CZ80IO() {};
		virtual void Write(CZ80 &CPU, Uint16 Addr, Uint8 Value8, Uint64 Value64, Uint32 Time) = 0;
		virtual void Read(CZ80 &CPU, Uint16 Addr, Uint8 &Value8, Uint64 &Value64, Uint32 Time) = 0;
};

class CZ80
{
	public:
		CZ80();
		~CZ80();

		void AllocateMemory(int slots);
		void Page(int slot, int readsrc, int writesrc);
		void GetPaged(int slot, int &read, int &write);
		void SetTiming(int slot, bool Contended);
		void SetGatherTimes(Uint32 CopyStartTime, Uint32 CopyPeriod, Uint32 CopyStride, Uint32 CopyLines);

		Uint32 Update(Uint32 cycles, bool);
		void SetIRQ(bool, char *dataptr = NULL, int instr = 0);

		RegisterState GetState();
		void SetState(RegisterState);

		Uint8 *GetPtr8(int slot);
		Uint64 *GetPtr64(int slot);
		Uint8 *GetContendedTable();
		void DumpMultiplexedCopy();

		void WriteMemory(Uint16 Addr, Uint8 Value8, Uint64 Value64);
		void WriteMemory(Uint16 Addr, Uint8 Value8);
		void ReadMemory(Uint16 Addr, Uint8 &Value8);
		void ReadMemory(Uint16 Addr, Uint8 &Value8, Uint64 &Value64);

		void WritePort(Uint16 Addr, Uint8 Value8, Uint64 Value64, unsigned int Timestamp);
		void ReadPort(Uint16 Addr, Uint8 &Value8, Uint64 &Value64, unsigned int Timestamp);
		void WritePort(Uint16 Addr, Uint8 Value8, unsigned int Timestamp);
		void ReadPort(Uint16 Addr, Uint8 &Value8, unsigned int Timestamp);

		void SetIOReadDevice(Uint16 Mask, Uint16 Value, CZ80IO *);
		void SetIOWriteDevice(Uint16 Mask, Uint16 Value, CZ80IO *);
		void SetIOContention(Uint16 Mask, Uint16 Value, bool);

		void SetTrapFunction( void (* Trap)(void *, CZ80 &, Uint16 Addr), void *param );
		void SetTrapAddr(Uint16 Addr, bool);

		void SetCopyAddrs(int *);
		void SetCopyMask(int);
		void SetCopyTarget(Uint8 *, Uint64 *);
	private:
		/* interrupt status */
		bool IFF1, IFF2;
		Uint8 IMode, IReg;
		bool ILine;

		/* refresh */
		Uint8 RReg, RHigh;

		/* normal registers */
		Uint8 AReg;
		RegPair BC, DE, HL, SP, BC2, DE2, HL2, AF2, IY, IX, PC;

		/* wide registers */
		Uint64 ABig, ABig2;
		Uint64 BBig, CBig, BBig2, CBig2;
		Uint64 DBig, EBig, DBig2, EBig2;
		Uint64 HBig, LBig, HBig2, LBig2;
		Uint64 IXlBig, IXhBig, IYlBig, IYhBig;

		/* processor status flags */
		bool LastOp;
		Uint8 Zero, Sign, HalfCarry, Carry, Misc, Overflow;
		Uint64 CarryBig, CarryBig2;

		/* memory mapper */
		Uint8 *Read8Ptrs[4], *Write8Ptrs[4];
		Uint64 *Read64Ptrs[4], *Write64Ptrs[4];

		Uint8 *MemoryPool8;
		Uint64 *MemoryPool64;
		int MemorySize;

		/* timing */
		Uint32 CyclesToRun;
		Uint32 TotalCycleCount, SubCycleCount, CopyTime;
		Uint8 TimePool[2*70000];
		Uint8 *TimePtrs[4];
		Uint8 **TimeTypes;

		/* copy parameters */
		int *CopySource;
		int CopySourceMask;
		Uint8 *TargetBuffer8;
		Uint64 *TargetBuffer64;

		Uint32 CopyStartTime, CopyPeriod, CopyStride, CopyLines;

		/* tables & misc */
		static Uint8 ParityTab[256];

		Uint32 CatchTab[2048];
		void (* TrapFunc)(void *, CZ80 &, Uint16 Addr);
		void *TrapParam;

		CZ80IO *IOReadDevice[65536], *IOWriteDevice[65536];
		Uint32 IOContention[2048];

		SDL_sem *GoSemaphore, *StopSemaphore;
		SDL_Thread *MainEx;
		static int CPUThreadHelper(void *tptr);
		int CPUThread();
		volatile bool Quit;
		void DoBreak();

		int FrameOffset, Frames;
		int BaseFrame;
};

#endif
