#include "Z80.h"
#include <memory.h>
#include <stdlib.h>

#define Read8Free(n)		Read8Ptrs[((n)&0xffff) >> 14][((n)&0xffff)&16383]
#define Read64Free(n)		Read64Ptrs[((n)&0xffff) >> 14][((n)&0xffff)&16383]
#define Write8Free(n, v)	Write8Ptrs[((n)&0xffff) >> 14][((n)&0xffff)&16383] = v
#define Write64Free(n, v, rv)	Write8Ptrs[((n)&0xffff) >> 14][((n)&0xffff)&16383] = v; Write64Ptrs[((n)&0xffff) >> 14][((n)&0xffff)&16383] = rv;

void CZ80::WritePort(Uint16 Addr, Uint8 Value8, Uint64 Value64, unsigned int Timestamp)
{
	if(IOWriteDevice[Addr]) IOWriteDevice[Addr]->Write(*this, Addr, Value8, Value64, Timestamp);
}

void CZ80::ReadPort(Uint16 Addr, Uint8 &Value8, Uint64 &Value64, unsigned int Timestamp)
{
	if(IOReadDevice[Addr]) IOReadDevice[Addr]->Read(*this, Addr, Value8, Value64, Timestamp);
	else
	{
		if((CopyTime > CopyStartTime) && (CopyTime < CopyStartTime + CopyLines*CopyStride) && ((CopyTime - CopyStartTime)%CopyStride < CopyPeriod))
		{
			Value8 = TargetBuffer8[(CopyTime-2) >>1];
			Value64 = TargetBuffer64[(CopyTime-2) >>1];
		}
		else
		{
			Value8 = 0xff;
			Value64 = 0xffffffffffffffffull;
		}
	}
}

void CZ80::SetGatherTimes(Uint32 xCopyStartTime, Uint32 xCopyPeriod, Uint32 xCopyStride, Uint32 xCopyLines)
{
	CopyStartTime = xCopyStartTime;
	CopyPeriod = xCopyPeriod;
	CopyStride = xCopyStride;
	CopyLines = xCopyLines;
}

void CZ80::WritePort(Uint16 Addr, Uint8 Value8, unsigned int Timestamp)
{
	WritePort(Addr, Value8, 0, Timestamp);
}

void CZ80::ReadPort(Uint16 Addr, Uint8 &Value8, unsigned int Timestamp)
{
	Uint64 Throwaway = 0;
	ReadPort(Addr, Value8, Throwaway, Timestamp);
}

CZ80::CZ80()
{
	GoSemaphore = SDL_CreateSemaphore(0);
	StopSemaphore = SDL_CreateSemaphore(0);

	MainEx = SDL_CreateThread(CPUThreadHelper, this);

	int c = 256;
	while(c--)
	{
		ParityTab[c] = c^1;
		ParityTab[c] = (ParityTab[c] >> 4) ^ (ParityTab[c] & 15);
		ParityTab[c] = (ParityTab[c] >> 2) ^ (ParityTab[c] & 3);
		ParityTab[c] = (ParityTab[c] >> 1) ^ (ParityTab[c] & 1);
		ParityTab[c] <<= 7;
	}

	MemoryPool8 = NULL;
	MemoryPool64 = NULL;

	Quit = false;
	SubCycleCount = CopyTime = TotalCycleCount = 0;
	ILine = false;

	PC.a = 0;
	IMode = 0;

	c = 65536;
	while(c--)
		IOReadDevice[c] = IOWriteDevice[c] = NULL;

	memset(TimePool, 0, 2*70000);
	memset(CatchTab, 0, 2048*sizeof(Uint32));
	memset(IOContention, 0, 2048*sizeof(Uint32));

	CopySourceMask = 0;
	CarryBig = 0;
}

void CZ80::SetIOReadDevice(Uint16 Mask, Uint16 Value, CZ80IO *d)
{
	int Addr = 65536;
	while(Addr--)
	{
		if((Addr&Mask) == Value)
			IOReadDevice[Addr] = d;
	}
}

void CZ80::SetIOWriteDevice(Uint16 Mask, Uint16 Value, CZ80IO *d)
{
	int Addr = 65536;
	while(Addr--)
	{
		if((Addr&Mask) == Value)
			IOWriteDevice[Addr] = d;
	}
}

void CZ80::SetIOContention(Uint16 Mask, Uint16 Value, bool t)
{
	int Addr = 65536;
	while(Addr--)
	{
		if((Addr&Mask) == Value)
		{
			if(t)
				IOContention[Addr >> 5] |= 1 << (Addr&31);
			else
				IOContention[Addr >> 5] &= ~(1 << (Addr&31));
		}
	}
}

CZ80::~CZ80()
{
	Quit = true;
	CyclesToRun = 50*256;
	SDL_SemPost(GoSemaphore);
	SDL_SemWait(StopSemaphore);

	int TResult;
	SDL_WaitThread(MainEx, &TResult);
	SDL_DestroySemaphore(GoSemaphore);
	SDL_DestroySemaphore(StopSemaphore);

	delete[] MemoryPool8;
	delete[] MemoryPool64;
}

void CZ80::SetIRQ(bool t, char *dataptr, int instr)
{
	ILine = t;
}

Uint32 CZ80::Update(Uint32 t, bool)
{
	if(t)
	{
		CyclesToRun = t;

		/* signal main thread to continue for a bit */
		SDL_SemPost(GoSemaphore);
		/* wait for it to finish */
		SDL_SemWait(StopSemaphore);
	}
	else
		CyclesToRun = 0;

	return 0;//CYCLENO_ANY;
}

void CZ80::AllocateMemory(int slots)
{
	MemorySize = slots << 14;
	MemoryPool8 = new Uint8[slots << 14];
	MemoryPool64 = new Uint64[slots << 14];
	TimeTypes = new Uint8*[slots];

	int c = slots << 14;
	while(c--)
	{
		MemoryPool64[c] = rand();
		MemoryPool64[c] += ((Uint64)rand()) << 16;
		MemoryPool64[c] += ((Uint64)rand()) << 32;
		MemoryPool64[c] += ((Uint64)rand()) << 48;
	}

	while(slots--)
		TimeTypes[slots] = TimePool;
}

void CZ80::SetTiming(int slot, bool Contended)
{
	TimeTypes[slot] = &TimePool[Contended ? 70000 : 0];
}

Uint8 *CZ80::GetContendedTable()
{
	return &TimePool[70000];
}

void CZ80::GetPaged(int slot, int &read, int &write)
{
	read = (int)(Read8Ptrs[slot] - MemoryPool8) >> 14;
	write = (int)(Write8Ptrs[slot] - MemoryPool8) >> 14;
}

void CZ80::Page(int slot, int readsrc, int writesrc)
{
	Read8Ptrs[slot] = &MemoryPool8[readsrc << 14];
	Read64Ptrs[slot] = &MemoryPool64[readsrc << 14];

	Write8Ptrs[slot] = &MemoryPool8[writesrc << 14];
	Write64Ptrs[slot] = &MemoryPool64[writesrc << 14];

	TimePtrs[slot] = TimeTypes[readsrc];
}

Uint8 *CZ80::GetPtr8(int src)
{
	return &MemoryPool8[src << 14];
}

Uint64 *CZ80::GetPtr64(int src)
{
	return &MemoryPool64[src << 14];
}

void CZ80::ReadMemory(Uint16 Addr, Uint8 &Value8)
{
	Value8 = Read8Free(Addr);
}

void CZ80::ReadMemory(Uint16 Addr, Uint8 &Value8, Uint64 &Value64)
{
	Value8 = Read8Free(Addr);
	Value64 = Read64Free(Addr);
}

void CZ80::WriteMemory(Uint16 Addr, Uint8 Value8, Uint64 Value64)
{
	Write64Free(Addr, Value8, Value64);
}

void CZ80::WriteMemory(Uint16 Addr, Uint8 Value8)
{
	WriteMemory(Addr, Value8,
				((Value8&0x80) ? 0xff00000000000000ull : 0) |
				((Value8&0x40) ? 0x00ff000000000000ull : 0) |
				((Value8&0x20) ? 0x0000ff0000000000ull : 0) |
				((Value8&0x10) ? 0x000000ff00000000ull : 0) |
				((Value8&0x08) ? 0x00000000ff000000ull : 0) |
				((Value8&0x04) ? 0x0000000000ff0000ull : 0) |
				((Value8&0x02) ? 0x000000000000ff00ull : 0) |
				((Value8&0x01) ? 0x00000000000000ffull : 0));
}

void CZ80::SetTrapFunction( void (* Trap)(void *, CZ80 &, Uint16 Addr), void *Param )
{
	TrapFunc = Trap;
	TrapParam = Param;
}

void CZ80::SetTrapAddr(Uint16 Addr, bool s)
{
	if(s)
		CatchTab[Addr >> 5] |= 1 << (Addr&31);
	else
		CatchTab[Addr >> 5] &= ~(1 << (Addr&31));
}

void CZ80::SetCopyAddrs(int *i)
{
	CopySource = i;
}

void CZ80::SetCopyMask(int m)
{
	CopySourceMask = m;
}

void CZ80::SetCopyTarget(Uint8 *ptr8, Uint64 *ptr64)
{
	TargetBuffer8 = ptr8;
	TargetBuffer64 = ptr64;
}

void CZ80::DumpMultiplexedCopy()
{
	int c = MemorySize;
	Uint8 *SPtr;
	Uint64 *DPtr;
	SPtr = GetPtr8(0);
	DPtr = GetPtr64(0);
	while(c--)
	{
		*DPtr =
				((*SPtr&0x80) ? 0xff00000000000000ull : 0) |
				((*SPtr&0x40) ? 0x00ff000000000000ull : 0) |
				((*SPtr&0x20) ? 0x0000ff0000000000ull : 0) |
				((*SPtr&0x10) ? 0x000000ff00000000ull : 0) |
				((*SPtr&0x08) ? 0x00000000ff000000ull : 0) |
				((*SPtr&0x04) ? 0x0000000000ff0000ull : 0) |
				((*SPtr&0x02) ? 0x000000000000ff00ull : 0) |
				((*SPtr&0x01) ? 0x00000000000000ffull : 0);

		DPtr++; SPtr++;
	}
}
